@@ -706,13 +706,9 @@ static gboolean real_emit_call_list_changed(void *data)
 
 static void emit_call_list_changed(struct ofono_voicecall *vc)
 {
-#ifdef DELAY_EMIT
 	if (vc->emit_calls_source == 0)
 		vc->emit_calls_source =
 			g_timeout_add(0, real_emit_call_list_changed, vc);
-#else
-	real_emit_call_list_changed(vc);
-#endif
 }
 
 static gboolean real_emit_multiparty_call_list_changed(void *data)
@@ -738,14 +734,9 @@ static gboolean real_emit_multiparty_call_list_changed(void *data)
 
 static void emit_multiparty_call_list_changed(struct ofono_voicecall *vc)
 {
-#ifdef DELAY_EMIT
 	if (vc->emit_multi_source == 0)
 		vc->emit_multi_source = g_timeout_add(0,
 				real_emit_multiparty_call_list_changed, vc);
-	}
-#else
-	real_emit_multiparty_call_list_changed(vc);
-#endif
 }
 
 static void voicecalls_release_queue(struct ofono_voicecall *vc, GSList *calls)
@@ -854,7 +845,7 @@ static struct ofono_call *synthesize_outgoing_call(struct ofono_voicecall *vc,
 	if (!call)
 		return call;
 
-	call->id = __ofono_modem_alloc_callid(modem);
+	call->id = __ofono_modem_callid_next(modem);
 
 	if (call->id == 0) {
 		ofono_error("Failed to alloc callid, too many calls");
@@ -862,6 +853,8 @@ static struct ofono_call *synthesize_outgoing_call(struct ofono_voicecall *vc,
 		return NULL;
 	}
 
+	__ofono_modem_callid_hold(modem, call->id);
+
 	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &number,
 				DBUS_TYPE_INVALID) == FALSE)
 		number = "";
@@ -1473,6 +1466,8 @@ void ofono_voicecall_disconnected(struct ofono_voicecall *vc, int id,
 
 	ofono_debug("Got disconnection event for id: %d, reason: %d", id, reason);
 
+	__ofono_modem_callid_release(modem, id);
+
 	l = g_slist_find_custom(vc->call_list, GUINT_TO_POINTER(id),
 				call_compare_by_id);
 
@@ -1504,8 +1499,6 @@ void ofono_voicecall_disconnected(struct ofono_voicecall *vc, int id,
 
 	vc->release_list = g_slist_remove(vc->release_list, call);
 
-	__ofono_modem_release_callid(modem, id);
-
 	if (reason != OFONO_DISCONNECT_REASON_UNKNOWN)
 		voicecall_emit_disconnect_reason(call, reason);
 
@@ -1550,6 +1543,8 @@ void ofono_voicecall_notify(struct ofono_voicecall *vc,
 
 	ofono_debug("Did not find a call with id: %d\n", call->id);
 
+	__ofono_modem_callid_hold(modem, call->id);
+
 	newcall = g_memdup(call, sizeof(struct ofono_call));
 
 	if (!newcall) {
@@ -1557,12 +1552,6 @@ void ofono_voicecall_notify(struct ofono_voicecall *vc,
 		goto err;
 	}
 
-	if (__ofono_modem_alloc_callid(modem) != call->id) {
-		ofono_error("Warning: Call id and internally tracked id"
-				" do not correspond");
-		goto err;
-	}
-
 	v = voicecall_create(vc, newcall);
 
 	if (!v) {
@@ -1887,3 +1876,14 @@ void *ofono_voicecall_get_data(struct ofono_voicecall *vc)
 {
 	return vc->driver_data;
 }
+
+int ofono_voicecall_get_next_callid(struct ofono_voicecall *vc)
+{
+	struct ofono_modem *modem;
+	if (vc == NULL || vc->atom == NULL)
+		return 0;
+
+	modem = __ofono_atom_get_modem(vc->atom);
+
+	return __ofono_modem_callid_next(modem);
+}
